--[[
  MTarget Version 0.2 ($Id: MTarget.lua 22 2009-02-11 12:37:20Z gameldar $)
  Written by gameldar
  http://www.wowace.com/projects/mtarget

  Thanks to:
    Ace3: Great lib framework
    ItemRack: Minimap Menu frame code is largely borrowed from here
    Recount: rules/handling for reseting group info
  ]]--

local version = {};
version.major = 0;
version.minor = 2;
version.build = "$Id: MTarget.lua 22 2009-02-11 12:37:20Z gameldar $";
version.version = version.major .. "." .. version.minor .. " (" .. version.build .. ")";
version.fullversion = "MTarget " .. version.version;

MTarget = LibStub("AceAddon-3.0"):NewAddon("MTarget", "AceConsole-3.0", "AceHook-3.0", "AceComm-3.0", "AceEvent-3.0", "AceSerializer-3.0")
local gui = LibStub("AceGUI-3.0")
local timer = LibStub("AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MTarget", true);

function MTarget:OnInitialize()
  -- initialize the MTarget addon when first loaded
  -- register the slash commands
  MTarget:RegisterChatCommand("mtargsetbyname", "SetTargetByName")
  MTarget:RegisterChatCommand("mtargsetbyunitid", "SetTargetByUnitId")
  MTarget:RegisterChatCommand("mtargsetvariable", "SetVariableString")
  MTarget:RegisterChatCommand("mtarglist", "PrintMapping")
  MTarget:RegisterChatCommand("mtarget", "MTargetControl")
  MTarget:RegisterChatCommand("mtargclearlist", "ClearMapping")
  MTarget:RegisterChatCommand("mtargcreatetemplate", "CreateTemplateFromMacro")
  MTarget:RegisterChatCommand("mtarglisttemplates", "PrintTemplateNames")
  MTarget:RegisterChatCommand("mtargshowtemplate", "PrintTemplate")
  MTarget:RegisterChatCommand("mtargdeletetemplate", "DeleteTemplate")
  MTarget:RegisterChatCommand("mtargwritemacros", "WriteMacros")
  MTarget:RegisterChatCommand("mtargbroadcast", "BroadcastVariable")
  MTarget:RegisterChatCommand("mtargresettemplates", "ResetTemplates")
  MTarget:RegisterChatCommand("mtargresetbroadcast", "ResetBroadcasts")
  MTarget:RegisterChatCommand("mtargupdateunitids", "UpdatePartyUnitIds")
  MTarget:Print(L["Initialized X"](version.version));
  MTarget.db = LibStub("AceDB-3.0"):New("MTargetDB")

  local options = {
	    name = "MTarget",
	    handler = MTarget,
	    hidden = false,
	    type = 'group',
	    args = {
		autowrite = {
		    type = 'toggle',
		    name = L['Auto Write Macros'],
		    desc = L['Rewrite the macros when a variable is changed.'],
		    set = 'SetAutoWrite',
		    get = 'GetAutoWrite',
		},
		showtooltips = {
		    type = 'toggle',
		    name = L['Show Tooltips'],
		    desc = L['Show tooltips on menu items.'],
		    set = 'SetShowTooltips',
		    get = 'GetShowTooltips',
		},
		showminimap = {
		    type = 'toggle',
		    name = L['Show Minimap'],
		    desc = L['Show Minimap button'],
		    set = 'SetShowMinimap',
		    get = 'GetShowMinimap',
		},
		squareminimap = {
		    type = 'toggle',
		    name = L['Square Minimap'],
		    desc = L['Minimap is Square'],
		    set = 'SetSquareMinimap',
		    get = 'GetSquareMinimap',
		},
		broadcast = {
		    type = 'toggle',
		    name = L['Broadcast variables'],
		    desc = L['Broadcast variables to party or raid'],
		    set = 'SetBroadcast',
		    get = 'GetBroadcast',
		},
    autoupdateunitid = {
        type = 'toggle',
        name = L['Auto update on party/raid changes'],
        desc = L['Update variable assignments when a party/raid changes composition or positions'],
        set = 'SetAutoUpdateUnitId',
        get = 'GetAutoUpdateUnitId',
    },
		autoaccept = {
		    type = 'toggle',
		    name = L['Auto accept Broadcasts'],
		    desc = L['Always accept broadcast from party/raid members'],
		    set = 'SetAutoAccept',
		    get = 'GetAutoAccept',
		},
		raidtargets = {
		    type = 'toggle',
		    name = L['Assign Raid Icons'],
		    desc = L['Automatically assign raid icons to varibles when you set them ("$star","$circle","$diamond","$triangle","$moon","$square","$cross","$skull")'],
		    set = 'SetRaidTargets',
		    get = 'GetRaidTargets',
		},
		debug = {
		    type = 'toggle',
		    name = L['Debug'],
		    desc = L['Enable debug output'],
		    set = 'SetDebug',
		    get = 'GetDebug',
		},
	    },
	}
  config = LibStub("AceConfig-3.0");
  aceconfigdialog = LibStub("AceConfigDialog-3.0");
  config:RegisterOptionsTable("MTarget", options, {"mtargconf"});
  aceconfigdialog:SetDefaultSize("MTarget", 1000, 600);

  config:RegisterOptionsTable( "MTarget", options )
  aceconfigdialog:AddToBlizOptions( "MTarget", "MTarget")

  UnitPopupButtons["TARGETER"] = {text = L["Set Target Variable..."], dist = 0, tooltipText = L["Set targeting variables."]};
  tinsert(UnitPopupMenus["PLAYER"], #UnitPopupMenus["PLAYER"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["FRIEND"], #UnitPopupMenus["FRIEND"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["TARGET"], #UnitPopupMenus["TARGET"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["SELF"], #UnitPopupMenus["SELF"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["PET"], #UnitPopupMenus["PET"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["PARTY"], #UnitPopupMenus["PARTY"] - 1, "TARGETER");
  tinsert(UnitPopupMenus["RAID"], #UnitPopupMenus["RAID"] - 1, "TARGETER");
  MTarget:SecureHook("UnitPopup_OnClick");
  MTarget:SecureHook("UnitPopup_ShowMenu");
  MTarget:CreateDialog();
  MTarget:InitRaidIcon();
  MTarget.menuOrient = "VERTICAL";
end
  MTarget.menuOrient = "VERTICAL";

MTarget.DockInfo = {  -- docking-dependent values
  LEFT = { xoff=1, yoff=0, menuSide="TOP", menuDir="TOP", orient="VERT", xadd=40, yadd=0 },
  RIGHT = { xoff=-1, yoff=0, menuSide="TOP", menuDir="TOP", orient="VERT", xadd=40, yadd=0 },
  TOP = { xoff=0, yoff=-1, menuSide="LEFT", menuDir="LEFT", orient="HORZ", xadd=0, yadd=40 },
  BOTTOM = { xoff=0, yoff=1, menuSide="LEFT", menuDir="LEFT", orient="HORZ", xadd=0, yadd=40 },
  TOPRIGHTTOPLEFT = { xoff=0, yoff=8,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 },
  BOTTOMRIGHTBOTTOMLEFT = { xoff=0, yoff=-8,  xdir=1,  ydir=1,  xstart=8,   ystart=44 },
  TOPLEFTTOPRIGHT = { xoff=0,  yoff=8,  xdir=-1, ydir=-1, xstart=-44, ystart=-8 },
  BOTTOMLEFTBOTTOMRIGHT = { xoff=0,  yoff=-8,  xdir=-1, ydir=1,  xstart=-44, ystart=44 },
  TOPRIGHTBOTTOMRIGHT = { xoff=8,  yoff=0, xdir=-1, ydir=1,  xstart=-44,  ystart=44 },
  BOTTOMRIGHTTOPRIGHT = { xoff=8,  yoff=0, xdir=-1, ydir=-1, xstart=-44,  ystart=-8 },
  TOPLEFTBOTTOMLEFT = { xoff=-8,  yoff=0, xdir=1,  ydir=1,  xstart=8,   ystart=44 },
  BOTTOMLEFTTOPLEFT = { xoff=-8,  yoff=0,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 },
}


function MTarget:InitAcceptList()
  MTarget:PartyMembersChanged();
  MTarget:RegisterEvent("PARTY_MEMBERS_CHANGED","PartyMembersChanged")
  MTarget:RegisterEvent("RAID_ROSTER_UPDATE","PartyMembersChanged")
end

function MTarget:InitRaidIcon()
  MTarget.raidtargets = {
    "star",
    "circle",
    "diamond",
    "triangle",
    "moon",
    "square",
    "cross",
    "skull",
  };
  MTarget:SecureHook("SetRaidTarget");
end

function MTarget:SetRaidTarget(unit, index)
  if (MTarget.db.char.options.raidtargets and MTarget.raidtargets[index]) then
    local map = MTarget:GetTargetUnitId(unit);
    if (map) then
      MTarget:SetTargetByUnitId(MTarget.raidtargets[index], map);
    else
      map = {
        target = UnitName(unit),
        guid = UnitGUID(unit),
        targettype = "name",
      };
      MTarget:SetTargetByName(MTarget.raidtargets[index], map);
    end
  end
end

function MTarget:PartyMembersChanged()
  local numRaidMembers = GetNumRaidMembers();
  local numPartyMembers = GetNumPartyMembers();

  if (numPartyMembers > 0 or numRaidMembers > 0) then
    if (not MTarget.inGroup) then
      MTarget.acceptlist = {};
    end
    MTarget.inGroup = true;
  else
    MTarget.acceptlist = {};
    MTarget.inGroup = false;
  end

  -- if autoupdateunitid then update the party unit ids
  if (MTarget.db.char.options.autoupdateunitid) then
    MTarget:UpdatePartyUnitIds();
  end
end

function MTarget:UpdatePartyUnitIds(input)
  changed = false;
  for name,value in pairs(MTarget.db.char.namemap) do
    if (value.targettype == "unitid") then
      targ = MTarget:GetUnitIdFromGUID(value.guid);
      if (targ ~= value.target) then
        changed = true;
        if (targ) then
          value.target = targ;
        else
          MTarget.db.char.namemap[name] = nil;
        end
      end
    end
  end

  if (changed) then
    MTarget:AutoWriteAndBroadcast(nil, nil);
    MTarget:BroadcastVariable();
  end
end

function MTarget:SetAutoWrite(info, input)
  MTarget.db.char.options.autowrite = input
end
function MTarget:GetAutoWrite(info)
  return MTarget.db.char.options.autowrite
end
function MTarget:SetShowTooltips(info, input)
  MTarget.db.char.options.showtooltips = input
end
function MTarget:GetShowTooltips(info)
  return MTarget.db.char.options.showtooltips
end
function MTarget:SetShowMinimap(info, input)
  MTarget.db.char.options.showminimap = input
  if (MTarget.db.char.options.showminimap) then
    MTargetMinimapFrame:Show()
  else
    MTargetMinimapFrame:Hide()
  end
end
function MTarget:GetShowMinimap(info)
  if (MTarget.db.char.options.showminimap == nil) then
    MTarget.db.char.options.showminimap = true;
  end
  return MTarget.db.char.options.showminimap
end
function MTarget:SetSquareMinimap(info, input)
  MTarget.db.char.options.squareminimap = input
end
function MTarget:GetSquareMinimap(info)
  return MTarget.db.char.options.squareminimap
end
function MTarget:SetBroadcast(info, input)
  MTarget.db.char.options.broadcast = input
end
function MTarget:GetBroadcast(info)
  return MTarget.db.char.options.broadcast
end
function MTarget:SetAutoAccept(info, input)
  MTarget.db.char.options.autoaccept = input
end
function MTarget:GetAutoAccept(info)
  return MTarget.db.char.options.autoaccept
end
function MTarget:SetRaidTargets(info, input)
  MTarget.db.char.options.raidtargets = input
end
function MTarget:GetRaidTargets(info)
  return MTarget.db.char.options.raidtargets
end
function MTarget:SetDebug(info, input)
  MTarget.db.char.argoptions.debug = input
end
function MTarget:GetDebug(info)
  return MTarget.db.char.options.debug
end
function MTarget:SetAutoUpdateUnitId(info, input)
  MTarget.db.char.options.autoupdateunitid = input
end
function MTarget:GetAutoUpdateUnitId(info)
  return MTarget.db.char.options.autoupdateunitid
end

local function trim(input)
  if (input == nil) then return; end;
  return string.gsub(input, "^(%s+)(.*)(%s+)$", "%2")
end

function MTarget:MTargetControl(input)
  input = trim(input);
  local b, e = string.find(input, "%s")
  local command;
  local args;
  if (b ~= nil) then
    command = trim(string.sub(input, 0, b - 1))
    args = trim(string.sub(input, e + 1))
    MTarget:Debug("command = " .. command .. "'");
    MTarget:Debug("args = '" .. args .. "'");
  elseif (input ~= "") then
    command = input
    args = nil
    MTarget:Debug("command = input = '" .. command .. "'");
    MTarget:Debug("args = nil");
  else
    command = nil
    args = nil
    MTarget:Debug("command = nil");
    MTarget:Debug("args = nil");
  end
  if (command == "setbyname") then
    MTarget:SetTargetByName(args);
  elseif (command == "setbyunitid") then
    MTarget:SetTargetByUnitId(args);
  elseif (command == "setvariable") then
   MTarget:SetVariableString(args);
  elseif (command == "list") then
    MTarget:PrintMapping(args);
  elseif (command == "clear") then
    MTarget:ClearMapping(args);
  elseif (command == "createtemplate") then
    MTarget:CreateTemplateFromMacro(args);
  elseif (command == "listtemplates") then
    MTarget:PrintTemplateNames(args);
  elseif (command == "showtemplate") then
    MTarget:PrintTemplate(args);
  elseif (command == "resettemplate") then
    MTarget:ResetTemplates(args);
  elseif (command == "deletetemplate") then
    MTarget:DeleteTemplate(args);
  elseif (command == "write") then
    MTarget:WriteMacros(args);
  elseif (command == "broadcast") then
    MTarget:BroadcastVariable(args);
  elseif (command == "resetbroadcast") then
    MTarget:ResetBroadcasts(args);
  elseif (command == "updateunitids") then
    MTarget:UpdatePartyUnitIds(args);
  elseif (command == nil or command == "help") then
    MTarget:PrintHelp();
  else
    MTarget:Print(L["Error: invalid command 'X'"](input));
    MTarget:PrintHelp();
  end
end

function MTarget:PrintHelp()
    MTarget:Print("/mtarget <" .. L["command"] .."> <" .. L["arguments"] .. ">");
    MTarget:Print("  " .. L["commands"] .. ":");
    MTarget:Print("    setbyname <variable> - " .. L["set the current target name to the specified variable"]);
    MTarget:Print("    setbyunitid <variable> - " .. L["set the current target unitid to the specified variable"]);
    MTarget:Print("    setvariable <variable>=<value> - " .. L["set the variable to the specified value"]);
    MTarget:Print("    list - " .. L["list all the currently assigned variables and values"]);
    MTarget:Print("    clear {<variable>} - " .. L["clear the assigned variable. Without an argument all the assigned variables are cleared."]);
    MTarget:Print("    createtemplate <macroname> - " .. L["create a template from the macro"]);
    MTarget:Print("    listtemplates - " .. L["list all the templates"]);
    MTarget:Print("    showtemplate - " .. L["show the contents of the template"]);
    MTarget:Print("    deletetemplate <macroname> - " .. L["delete the template"]);
    MTarget:Print("    resettemplate {<macroname>} - " .. L["set the macro to have the template text. Without an argument it will reset all macros with templates"]);
    MTarget:Print("    updateunitids - " .. L["update variable assignments to match party/raid composition or positions if changed"]);
    MTarget:Print("    write - " .. L["create the macros from all the templates filling in the variables"]);
    MTarget:Print("    broadcast {<variable>} - " .. L["broadcast your variable to your party/raid. Without an argument it broadcasts all variables"]);
    MTarget:Print("    resetbroadcast - " .. L["clear any broadcast variables (sender accept, and batch flags)"]);
    MTarget:Print("    help - " .. L["print help information"]);
end
function MTarget:UnitPopup_ShowMenu(dropdownMenu, which, unit, name, userData)
  local server = nil;
  local map = nil;
  -- Init variables
  if ( unit == "target") then
    map = MTarget:GetTargetUnitId();
  end
  if (not map and unit and unit ~= "target") then
    map = {};
    name, server = UnitName(unit);
    map.target = unit;
    map.guid = UnitGUID(unit);
    map.targettype = "unitid";
  elseif ( not map) then
    map = {};
    local n, s = UnitName("target");
    if ( n ) then
      name = n;
      server = s;
    end
    map = {};
    map.target = name;
    map.guid = UnitGUID("playertarget");
    map.targettype = "name";
  end
  if (map) then
    MTarget.currentTarget = map;
    MTarget.currentTarget.name = name or UnitName(unit);
    MTarget.currentTarget.server = server;
    MTarget:Debug(L["Set target X to: Y[Z]"](MTarget.currentTarget.name, MTarget.currentTarget.target, MTarget.currentTarget.targettype));
  else
    MTarget:Debug(L["No map for current menu"]);
    MTarget.currentTarget = nil;
  end
end

function MTarget:UnitPopup_OnClick(self)
  local button = self.value;
  if(button == "TARGETER") then
    local txt = nil;
    if (MTarget.currentTarget.targettype == "unitid") then
      txt = MTarget.currentTarget.name .. "[" .. MTarget.currentTarget.target .. "]";
    else
      txt = MTarget.currentTarget.name;
    end
    StaticPopup_Show("MTarget_SetVariable", txt);
  end
end

function MTarget:CreateDialog()
  StaticPopupDialogs["MTarget_SetVariable"] = {
    text = L["Set variable for %s"],
    button1 = L["Set by Name"],
    button3 = L["Set by UnitId"],
    button2 = L["Cancel"],
    hasEditBox = 1,
    whileDead = 1,
    hideOnEscape = 1,
    timeout = 0,
    DisplayButton3 = function()
      return (MTarget.currentTarget ~= nil and MTarget.currentTarget.targettype == "unitid");
    end,
    OnShow = function()
      getglobal(this:GetName().."EditBox"):SetText("");
    end,
    OnAccept = function()
      variable = getglobal(this:GetParent():GetName().."EditBox"):GetText();
      MTarget:SetTargetByName(variable, MTarget.currentTarget);
    end,
    OnAlt = function()
      variable = getglobal(this:GetParent():GetName().."EditBox"):GetText();
      MTarget:SetTargetByUnitId(variable, MTarget.currentTarget);
    end
  };
  StaticPopupDialogs["MTarget_SetVariableString"] = {
    text = L["Set variable name and value string (e.g. spell=Exorcism)"],
    button1 = L["Set"],
    button2 = L["Cancel"],
    hasEditBox = 1,
    whileDead = 1,
    hideOnEscape = 1,
    timeout = 0,
    OnShow = function()
      getglobal(this:GetName().."EditBox"):SetText("x=y");
    end,
    OnAccept = function()
      variable = getglobal(this:GetParent():GetName().."EditBox"):GetText();
      MTarget:SetVariableString(variable);
    end
  };
  StaticPopupDialogs["MTarget_AcceptBroadcast"] = {
    text = L["%s wants to set %s."],
    button1 = L["Accept"],
    button3 = L["Always"],
    button2 = L["Ignore"],
    whileDead = 1,
    hideOnEscape = 1,
    timeout = 0,
    OnAccept = function()
      MTarget:SetFromBroadcast(false);
    end,
    OnAlt = function()
      MTarget:SetFromBroadcast(true);
    end
  };
  StaticPopupDialogs["MTarget_ResetTemplateConfirm"] = {
    text = L["Reset macro to template value for: %s"],
    button1 = L["Reset"],
    button2 = L["Cancel"],
    whileDead = 1,
    hideOnEscape = 1,
    timeout = 0,
    OnAccept = function()
      MTarget:ResetTemplates(MTarget.resetTemplate, true);
    end
  };
  StaticPopupDialogs["MTarget_DeleteTemplateConfirm"] = {
    text = L["Delete template: %s"],
    button1 = L["Delete"],
    button2 = L["Cancel"],
    whileDead = 1,
    hideOnEscape = 1,
    timeout = 0,
    OnAccept = function()
      MTarget:DeleteTemplate(MTarget.deleteTemplate, true);
    end
  };
end
--[[
Blizzard_MacroFrame_OnShow = MacroFrame_OnShow

MacroFrame_OnShow = function()
  Blizzard_MacroFrame_OnShow();
  MTarget:InitMacroButton();
  MacroFrame_OnShow = Blizzard_MacroFrame_OnShow
end
]]--
function MTarget:OnEnable()
  -- enable MTarget
  if (not MTarget.db.char.namemap) then
    MTarget.db.char.namemap = {}
  end
  if (not MTarget.db.char.templates) then
    MTarget.db.char.templates = {}
  end
  if (not MTarget.db.char.options) then
    MTarget.db.char.options = {}
  end
  if (not MTarget.batchSend) then
    MTarget.batchSend = {};
  end
--  MTarget:InitMacroButtons();
  MTarget:Debug("InitTimers");
  MTarget:InitTimers();
--  MTarget:Debug("Hooking OnShow");
  MTarget:RegisterEvent("ADDON_LOADED","AddOnLoaded");
--  MTarget:Debug("Hooking OnShow");
  MTarget:InitMenu();
  if (not MTarget:GetShowMinimap()) then
    MTargetMinimapFrame:Hide()
  end
  MTarget:InitAcceptList();
  MTarget:Debug("Enabled");
end

function MTarget:AddOnLoaded(event, name)
  if (name == "Blizzard_MacroUI") then
    MTarget:InitMacroButton();
  end
end

function MTarget:OnDisable()
  -- clean up on disable
end


local function validateVariableName(input)
  -- trim off any white space
  input = trim(input)
  -- if the input contains any spaces throw an error
  if (string.find(input, "%s") ~= nil) then
    MTarget:Print(L["Error: invalid variable name 'X'. Cannot contain spaces."](input))
    return nil
  end
  -- if the input contains a $ character throw an error
  if (string.find(input, "$", 0, true) ~= nil) then
    MTarget:Print(L["Error: invalid variable name 'X'. Cannot contain $."](input))
    return nil
  end
  return input
end

local function targetExists(unitid)
  if (not unitid) then
    unitid = "playertarget";
  end
  local ret = UnitExists(unitid)
  if (not ret) then
    MTarget:Print(L["Error: no unit targeted"])
  end
  return ret
end

function MTarget:GetNameMap(input)
  if (not MTarget.db.char.namemap[input]) then
      MTarget.db.char.namemap[input] = {}
  end
  return MTarget.db.char.namemap[input]
end

function MTarget:BroadcastTarget(name, tmap)
  local targ;

  if (GetNumRaidMembers() > 0) then
    targ = "RAID";
  elseif (GetNumPartyMembers() > 0) then
    targ = "PARTY";
  end
  if (targ) then
    local msg = MTarget:Serialize(name, tmap);
    MTarget:SendCommMessage("MTarget", msg, targ);
  end
end

function MTarget:OnCommReceived(prefix, message, distribution, sender)
  if (prefix == "MTarget" and sender ~= UnitName("player")) then
    if (message == "batchstart") then
      MTarget.batchSend[sender] = true;
    elseif (message == "batchend") then
      MTarget.batchSend[sender] = nil;
      MTarget:AutoWriteAndBroadcast(nil, nil);
    else
      success, name, map = MTarget:Deserialize(message);
      if (not success) then
        MTarget:Print(L["Error: failed to deserialize broadcast message: 'X'"](message));
        return;
      end
      if (map.targettype == "unitid") then
        map.target = MTarget:GetUnitIdFromGUID(map.guid);
      end
      map.name = name;
      map.sender = sender;
      MTarget.currentTarget = map;

      if (MTarget.db.char.options.autoaccept or MTarget.acceptlist[sender]) then
        MTarget.db.char.namemap[name] = map;
        if (not MTarget.batchSend[sender]) then
          MTarget:Print(L["Broadcast Set $X='Y'"](name, map.target))
          MTarget:AutoWriteAndBroadcast(nil, nil);
        end
      else
        StaticPopup_Show("MTarget_AcceptBroadcast", sender, name .. "[" .. map.targettype .. "]" .. "='" .. map.target .."'");
      end
    end
  end
end

function MTarget:ShowSetVariableString()
  StaticPopup_Show("MTarget_SetVariableString");
end

function MTarget:SetVariableString(input)
  local var, value;
  local b, e = string.find(input, "=");
  if (not b) then
    MTarget:Print(L["Error: invalid Variable String format. Must be x=y"]);
    return;
  end
  var = trim(string.sub(input, 0, b - 1));
  value = string.sub(input, e + 1);

  local map = MTarget:GetNameMap(var);
  map.target = value;
  map.guid = nil;
  map.targettype = "string";
  MTarget:Print(L["Set $X='Y'"](var, map.target))
  MTarget:AutoWriteAndBroadcast(var, map);
end

function MTarget:SetFromBroadcast(remember)
  if (remember) then
    if (not MTarget.acceptlist) then
      MTarget.acceptlist = {};
    end
    MTarget.acceptlist[MTarget.currentTarget.sender] = true;
  end
  local map = MTarget:GetNameMap(MTarget.currentTarget.name);
  map.target = MTarget.currentTarget.target;
  map.targettype = MTarget.currentTarget.targettype;
  map.guid = MTarget.currentTarget.guid;

  MTarget:Print(L["Broadcast Set $X='Y'"](MTarget.currentTarget.name, map.target))
  MTarget:AutoWriteAndBroadcast(nil, nil);
end

function MTarget:AutoWriteAndBroadcast(input, map)
  if (MTarget.db.char.options.autowrite) then
    MTarget:WriteMacros();
  end
  if (input and MTarget.db.char.options.broadcast) then
    MTarget:BroadcastTarget(input, map);
  end
end

function MTarget:SetTargetByName(input, targ)
  -- set the target variable to the name of the unit
  local validInput = validateVariableName(input)
  if (validInput == nil) then
    return
  end
  local map = MTarget:GetNameMap(validInput)
  if (targ or targetExists()) then
    if (targ) then
      map.target = targ.name;
      map.guid = targ.guid;
      map.targettype = "name";
    else
      map.target = UnitName("playertarget")
      map.guid = UnitGUID("playertarget")
      map.targettype = "name"
    end
    MTarget:AutoWriteAndBroadcast(input, map);
    MTarget:Print(L["Set $X='Y'"](validInput, map.target))
  end
end

function MTarget:PrintMapping(input)
  -- print out the current name=value pairs
  for name,value in pairs(MTarget.db.char.namemap) do
    local t = value.target or "nil";
    local tt = value.targettype or "nil";
    MTarget:Print(name .. "[" .. tt .. "]=" .. t);
  end
end

function MTarget:ClearMapping(input)
  -- clear the list of namemaps
  MTarget.db.char.namemap = {}
end

function MTarget:CreateTemplateFromMacro(input)
  -- create a macro template from an existing macro
  local macroIndex = GetMacroIndexByName(input)
  if (macroIndex == nil) then
    MTarget:Print(L["Error: no such macro 'X'"](input))
    return
  end
  MTarget:CreateTemplateFromMacroIndex(macroIndex);
end
function MTarget:CreateTemplateFromMacroIndex(macroIndex)
  local macroText = GetMacroBody(macroIndex)
  local name, iconTexture, body, localVar = GetMacroInfo(macroIndex)
  MTarget.db.char.templates[name] = macroText
  MTarget:Print(L["Created template from macro[X]: Y"](macroIndex, name))
end

function MTarget:PrintTemplateNames(input)
  -- print out the name of all the templates
  MTarget:Print(L["Templates"] .. ": ")
  for name,value in pairs(MTarget.db.char.templates) do
    MTarget:Print(name)
  end
end

function MTarget:PrintTemplate(input)
  -- print out the contents of the template
  if (not MTarget.db.char.templates[input]) then
    MTarget:Print(L["Error: no such template 'X'"](input))
    return
  end
  MTarget:Print(L["Macro Template"] .. ": " .. input)
  MTarget:Print(MTarget.db.char.templates[input])
end

function MTarget:DeleteTemplate(input, confirmed)
  -- delete a template
  if (not MTarget.db.char.templates[input]) then
    MTarget:Print(L["Error: no such template 'X'"](input))
    return
  end
  if (confirmed) then
    MTarget.deleteTemplate = nil;
    MTarget.db.char.templates[input] = nil;
  else
    MTarget.deleteTemplate = input
    StaticPopup_Show("MTarget_DeleteTemplateConfirm", input);
  end
end

function MTarget:ResetTemplates(input, confirmed)
  -- restore the macro to the template value
  if (input and not MTarget.db.char.templates[input] and input ~= L["all"]) then
    MTarget:Print(L["Error: no such template 'X'"](input))
    return
  end
  if (confirmed) then
    MTarget.resetTemplate = nil;
    map = {};
    for name,value in pairs(MTarget.db.char.namemap) do
      map[name] = value;
      value.target = "$" .. name;
    end
    if (not input or input == L["all"]) then
      MTarget:WriteMacros(nil, namemap);
    else
      MTarget:WriteMacro(input, MTarget.db.char.templates[input], namemap);
    end
  else
    if (not input) then
      input = L["all"];
    end
    MTarget.resetTemplate = input
    StaticPopup_Show("MTarget_ResetTemplateConfirm", input);
  end
end

function MTarget:BroadcastVariable(input)
  -- broadcast the variable(s) to the party/raid
  local targ;

  if (GetNumRaidMembers() > 0) then
    targ = "RAID";
  elseif (GetNumPartyMembers() > 0) then
    targ = "PARTY";
  end
  if (targ) then
    if (input) then
      map = MTarget.db.char.namemap[input];
      if (map) then
        MTarget:BroadcastTarget(input, map);
      else
        MTarget:Print(L["Error: Variable 'X' not assigned"](input));
      end
    else
      MTarget:SendCommMessage("MTarget", "batchstart", targ);
      for name,value in pairs(MTarget.db.char.namemap) do
        MTarget:BroadcastTarget(name, value);
      end
      MTarget:SendCommMessage("MTarget", "batchend", targ);
    end
  end
end

function MTarget:ResetBroadcasts(input)
  -- reset the broadcast variables
  MTarget.acceptlist = {};
  MTarget.batchSend = {};
end

function MTarget:WriteMacros(input, namemap)
  -- write out the macros replacing all the values with the name/value mappings

  if (not namemap) then
    namemap = MTarget.db.char.namemap;
  end

  for macroName,macroText in pairs(MTarget.db.char.templates) do
    MTarget:WriteMacro(macroName, macroText, namemap);
  end
end

function MTarget:WriteMacro(macroName, macroText, namemap)
  -- can't write out macros when in combat
  if (UnitAffectingCombat("player")) then
    MTarget:Print(L["Error: Cannot write macros while in combat"])
    return
  end
  for name,value in pairs(namemap) do
    macroText = string.gsub(macroText, "$" .. name, value.target)
  end
  local macroIndex = GetMacroIndexByName(macroName)
  if (macroIndex ~= 0) then
    EditMacro(macroIndex, macroName, 1, macroText, 1, 1)
    MTarget:Print(L["Updated Macro X"](macroName))
  else
    CreateMacro(macroName, 1, macroText, 1, 1)
    MTarget:Print(L["Created Macro X"](macroName))
  end
end

function MTarget:GetUnitIdFromGUID(guid)
  if (UnitGUID("player") == guid) then
    return "player";
  elseif (UnitGUID("pet") == guid) then
    return "pet";
  elseif (GetNumPartyMembers() > 0) then
    for i = 1, GetNumPartyMembers() - 1 do
      local unit = "party" .. i;
      local unitpet = "partypet" .. i;
      if (UnitGUID(unit) == guid) then
        return unit;
      elseif (UnitGUID(unitpet) == guid) then
        return unitpet;
      end
    end
  elseif (GetNumRaidMembers() > 0) then
    for i = 1, GetNumRaidMembers() - 1 do
      local unit = "raid" .. i;
      local unitpet = "raidpet" .. i;
      if (UnitGUID(unit) == guid) then
        return unit;
      elseif (UnitGUID(unitpet) == guid) then
        return unitpet;
      end
    end
  end
  return nil;
end

function MTarget:GetTargetUnitId(unitid)
  if (not unitid) then
    unitid = "playertarget";
  end
  -- get the target's raid/party id
  -- this only works on raid/party members (and their pets)
  local map = {};
  if (targetExists(unitid)) then
    if (UnitPlayerOrPetInRaid(unitid)) then
      for i = 1, GetNumRaidMembers() - 1 do
        local unit = "raid" .. i
        local unitpet = "raidpet" .. i
        if (UnitIsUnit(unit, unitid)) then
          map.target = unit
          map.guid = UnitGUID(unitid)
          map.targettype = "unitid"
          break
        elseif (UnitItUnit(unitpet, unitid)) then
          map.target = unitpet
          map.guid = UnitGUID(unitid)
          map.targettype = "unitid"
          break
        end
      end
    elseif (UnitPlayerOrPetInParty(unitid)) then
      for i = 1, GetNumPartyMembers() - 1 do
        local unit = "party" .. i
        local unitpet = "partypet" .. i
        if (UnitIsUnit(unit, unitid)) then
          map.target = unit
          map.guid = UnitGUID(unitid)
          map.targettype = "unitid"
          break
        elseif (UnitIsUnit(unitpet, unitid)) then
          map.target = unitpet
          map.guid = UnitGUID(unitid)
          map.targettype = "unitid"
          break
        end
      end
    elseif (UnitIsUnit("pet", unitid)) then
      map.target = "pet"
      map.guid = UnitGUID(unitid)
      map.targettype = "unitid"
    elseif (UnitIsUnit("player", unitid)) then
      map.target = "player"
      map.guid = UnitGUID(unitid)
      map.targettype = "unitid"
    else
      return nil;
    end
    return map;
  end
  return nil;
end

function MTarget:SetTargetByUnitId(input, targ)
  -- set the target by getting its raid/party id
  -- this only works on raid/party members
  local validInput = validateVariableName(input)
  if (validInput == nil) then
    return
  end
  local map;

  if (targ) then
    map = targ;
  else
    map = MTarget:GetTargetUnitId();
  end
  if (map) then
    local globalmap = MTarget:GetNameMap(validInput)
    globalmap.target = map.target;
    globalmap.guid = map.guid;
    globalmap.targettype = "unitid";
    MTarget:Print(L["Set $X='Y'"](validInput, MTarget.db.char.namemap[validInput].target))
    MTarget:AutoWriteAndBroadcast(validInput, map);
  else
    MTarget:Print(L["Error: target is not part of a group or raid"]);
  end
end

function MTarget:MacroFrame_Update()
  if (MacroFrame.selectedMacro ~= nil) then
    MTarget.createTemplateButton:Enable();
  else
    MTarget.createTemplateButton:Disable();
  end
end
function MTarget:CreateTemplateButtonOnClick(targ, button)
  if (MacroFrame.selectedMacro ~= nil) then
    MTarget:CreateTemplateFromMacroIndex(MacroFrame.selectedMacro);
  end
end


-- Minimap Button --

function MTarget:OnTooltip(line1,line2)
	if MTarget.db.char.options.showtooltips then
		MTarget:AnchorTooltip()
		if line1 then
			GameTooltip:AddLine(line1)
			GameTooltip:AddLine(line2,.8,.8,.8,1)
			GameTooltip:Show()
			return
		end
	end
end

function MTarget:AnchorTooltip(owner)
  GameTooltip_SetDefaultAnchor(GameTooltip,owner)
end

function MTarget:MinimapDragging()
	local xpos,ypos = GetCursorPosition()
	local xmin,ymin = Minimap:GetLeft(), Minimap:GetBottom()

	xpos = xmin-xpos/Minimap:GetEffectiveScale()+70
	ypos = ypos/Minimap:GetEffectiveScale()-ymin-70

	MTarget.db.char.options.iconpos = math.deg(math.atan2(ypos,xpos))
	MTarget:MoveMinimap()
end

function MTarget:MenuTooltip()
--	local id = this.id
--	MTarget:SetTooltip(MTarget.menu[id]);
end

function MTarget:MoveMinimap()
	if MTarget.db.char.options.showminimap then
		local xpos,ypos
		local angle = MTarget.db.char.options.iconpos or -100
		if MTarget.db.char.options.squareminimap then
			-- brute force method until trig solution figured out - min/max a point on a circle beyond square
			xpos = 110 * cos(angle)
			ypos = 110 * sin(angle)
			xpos = math.max(-82,math.min(xpos,84))
			ypos = math.max(-86,math.min(ypos,82))
		else
			xpos = 80 * cos(angle)
			ypos = 80 * sin(angle)
		end
		MTargetMinimapFrame:SetPoint("TOPLEFT","Minimap","TOPLEFT",52-xpos,ypos-52)
		MTargetMinimapFrame:Show()
	else
		MTargetMinimapFrame:Hide()
	end
end

function MTarget:DockWindows(menuDock, relativeTo, mainDock, menuOrient, movable)
  MTargetMenuFrame:ClearAllPoints()
  MTarget.currentDock = mainDock..menuDock
  MTargetMenuFrame:SetPoint(menuDock,relativeTo,mainDock,MTarget.DockInfo[MTarget.currentDock].xoff,MTarget.DockInfo[MTarget.currentDock].yoff)
  MTargetMenuFrame:SetParent(relativeTo)
  MTargetMenuFrame:SetFrameStrata("HIGH")
  MTarget.menuDock = menuDock
end

function MTarget:MinimapOnClick()
  if MTargetMenuFrame:IsVisible() then
    MTargetMenuFrame:Hide()
  else
    local xpos,ypos = GetCursorPosition();
    if ypos>400 then
      MTarget:DockWindows("TOPRIGHT",MTargetMinimapFrame,"BOTTOMRIGHT","VERTICAL");
    else
      MTarget:DockWindows("BOTTOMRIGHT",MTargetMinimapFrame,"TOPLEFT","VERTICAL");
    end
    MTarget:BuildMenu();
  end
end

function MTarget:MinimapOnEnter()
  if (MTarget.db.char.options.showtooltips) then
    MTarget:OnTooltip("MTarget","Click to show MTarget actions");
  end
end

function MTarget:CreateMenuItem(id, text, tooltip, callbackfunc)
  button = getglobal("MTargetMenu"..id);
  if (not button) then
--    MTarget.menu[id] = tooltip;
    button = CreateFrame("Button","MTargetMenu"..id,getglobal("MTargetMenuFrame"), "UIDropDownMenuButtonTemplate");
--[[
    button.normalText = button:CreateFontString(nil, "OVERLAY", "GameFontHighlightSmall");
    button.normalText:SetPoint("LEFT", button, "LEFT", -5, 0)
    button.normalText:SetJustifyH("LEFT")
    button:SetFontString(button.normalText)

    button:SetNormalFontObject(GameFontHighlightSmall)
    button:SetHighlightFontObject(GameFontHighlightSmall)
    button:SetDisabledFontObject(GameFontDisableSmall)
    button:SetFrameStrata("HIGH")
]]--
    getglobal(button:GetName().."Check"):Hide();

    button:SetText(text);
    button:RegisterForClicks("LeftButtonUp","RightButtonUp");
    button:SetScript("OnClick",function ()
      MTargetMenuFrame:Hide()
      MTarget[callbackfunc]();
      end);
--    button:SetScript("OnEnter",MTarget["MenuTooltip"]);
--    button:SetScript("OnLeave",MTarget["ClearTooltip"]);

--		CreateFrame("Frame",nil,button,"MTargetTimeTemplate")
--		MTarget:SetFont("MTargetMenu"..idx)

    end
  return button;
end

function MTarget:InitMenu()
  --if (not MTarget.menu) then
    MTarget:Debug("Initialising Menu");
    MTarget.menu = {}
    MTarget.menu[1] = MTarget:CreateMenuItem("setvariable", L["Set Variable..."], L["Set a variable to an arbitrary string value"], "ShowSetVariableString");
    MTarget.menu[2] = MTarget:CreateMenuItem("list", L["List Variables"], L["Print out the currently set variables."], "PrintMapping");
    MTarget.menu[3] = MTarget:CreateMenuItem("clear", L["Clear Variables"], L["Clear all currently set variables."], "ClearMapping");
    MTarget.menu[4] = MTarget:CreateMenuItem("templnames", L["List Templates"], L["List all the templates."], "PrintTemplateNames");
    MTarget.menu[5] = MTarget:CreateMenuItem("resettempl", L["Reset Templates"], L["Reset the macros to the template values."], "ResetTemplates");
    MTarget.menu[6] = MTarget:CreateMenuItem("updateunitids", L["Update Unit Ids"], L["update variable assignments to match party/raid composition or positions if changed"], "UpdatePartyUnitIds");
    MTarget.menu[7] = MTarget:CreateMenuItem("writemacros", L["Write Macros"], L["Write out all the macros from the templates with the currently set of variables."], "WriteMacros");
    MTarget.menu[8] = MTarget:CreateMenuItem("broadcast", L["Broadcast Variables"], L["Broadcast variables to all party/raid members"], "BroadcastVariable");
    MTarget.menu[9] = MTarget:CreateMenuItem("resetbc", L["Reset Broadcast"], L["Reset the broadcast variables (accept list, batch flag)"], "ResetBroadcasts");
  --end
    MTarget:Debug("Menu Initialized: " .. #(MTarget.menu));
end

function MTarget:BuildMenu()
  if (#MTarget.menu == 0) then
    MTarget:InitMenu();
  end
  local col,row,xpos,ypos = 0,0,MTarget.DockInfo[MTarget.currentDock].xstart,MTarget.DockInfo[MTarget.currentDock].ystart
  local max_cols = 1
  local button

  for i=1,#(MTarget.menu) do
    button = MTarget.menu[i];
    button:SetID(i);
    button:SetPoint("TOPLEFT",MTargetMenuFrame,"TOPLEFT",10,ypos)
    button:SetFrameLevel(MTargetMenuFrame:GetFrameLevel()+1)
--    if MTarget.menuOrient=="VERTICAL" then
--      xpos = xpos + MTarget.DockInfo[MTarget.currentDock].xdir*40
      col = col + 1
      if col==max_cols then
        xpos = MTarget.DockInfo[MTarget.currentDock].xstart
        col = 0
        ypos = ypos + MTarget.DockInfo[MTarget.currentDock].ydir*20
        row = row + 1
      end
      button:Show()
--[[    else
      ypos = ypos + MTarget.DockInfo[MTarget.currentDock].ydir*40
      col = col + 1
      if col==max_cols then
        ypos = MTarget.DockInfo[MTarget.currentDock].ystart
        col = 0
        xpos = xpos + MTarget.DockInfo[MTarget.currentDock].xdir*40
        row = row + 1
      end
      button:Show()

    end
]]--
  end

  if col==0 then
    row = row-1
  end

  if MTarget.menuOrient=="VERTICAL" then
    MTargetMenuFrame:SetWidth(100+(max_cols*20))
    MTargetMenuFrame:SetHeight(100+((row+1)*20))
  else
    MTargetMenuFrame:SetWidth(100+((row+1)*20))
    MTargetMenuFrame:SetHeight(100+(max_cols*20))
  end

  MTargetMenuFrame:Show();
end

function MTarget:InitMacroButton()
  macroDelete = getglobal("MacroDeleteButton");
  MTarget.createTemplateButton = CreateFrame("Button", "MTarget_CreateTemplateButton", macroDelete:GetParent(), "UIPanelButtonGrayTemplate");
  MTarget.createTemplateButton:SetText(L["Template"]);
  MTarget.createTemplateButton:SetScript("OnClick", MTarget["CreateTemplateButtonOnClick"]);
  MTarget.createTemplateButton:SetHeight(22);
  MTarget.createTemplateButton:SetWidth(80);
  MTarget.createTemplateButton:ClearAllPoints();
  MTarget.createTemplateButton:SetPoint("BOTTOMLEFT", "MacroFrame", "BOTTOMLEFT", 97, 79);
  MTarget:SecureHook("MacroFrame_Update");
end

-- Timers --
function MTarget:CreateTimer(name, callback, delay, rep, args)
  if (not MTarget.TimerPool[name]) then
    MTarget.TimerPool[name] = {};
  end
  MTarget.TimerPool[name].callback = callback;
  MTarget.TimerPool[name].delay = delay;
  MTarget.TimerPool[name].rep = rep;
  MTarget.TimerPool[name].args = args;
end
function MTarget:InitTimers()
  MTarget.TimerPool = {};
	MTarget:CreateTimer("MinimapDragging",MTarget.MinimapDragging,0, 1)
--	MTarget:CreateTimer("MinimapShine",MTarget:MinimapShineUpdate,0,1)
end


function MTarget:StartTimer(name)
  t = MTarget.TimerPool[name];
  if (t.rep) then
    t.handle = timer:ScheduleTimer( t.callback, t.delay, t.args );
  else
    t.handle = timer:ScheduleTimer( t.callback, t.delay, t.args );
  end
end

function MTarget:StopTimer(name)
  t = MTarget.TimerPool[name];
  if (t and t.handle) then
    timer:CancelTimer(t.handle);
  end
end

function MTarget:Debug(msg)
  if (MTarget.db.char.options.debug) then
    MTarget:Print(msg);
  end
end
